import logging

import numpy as np

from pymchelper.shieldhit.detector.detector_type import SHDetType
from pymchelper.shieldhit.detector.estimator_type import SHGeoType
from pymchelper.flair.Data import Usrbin, unpackArray

logger = logging.getLogger(__name__)


class FlukaBinaryReader:
    def __init__(self, filename):
        self.filename = filename

    def read(self, detector, nscale=1):
        usr = Usrbin(self.filename)
        usr.say()  # file,title,time,weight,ncase,nbatch
        for i, _ in enumerate(usr.detector):
            logger.debug("-" * 20 + (" Detector number %i " % i) + "-" * 20)
            usr.say(i)  # details for each detector
        data = usr.readData(0)
        fdata = unpackArray(data)

        # TODO read detector type
        detector.det = "FLUKA"

        # TODO read particle type
        detector.particle = 0

        # TODO read geo type
        detector.geotyp = SHGeoType.unknown

        # TODO cross-check statistics
        detector.nstat = usr.ncase

        # TODO figure out when more detectors are used
        detector.nx = usr.detector[0].nx
        detector.ny = usr.detector[0].ny
        detector.nz = usr.detector[0].nz

        detector.xmin = usr.detector[0].xlow
        detector.ymin = usr.detector[0].ylow
        detector.zmin = usr.detector[0].zlow

        detector.xmax = usr.detector[0].xhigh
        detector.ymax = usr.detector[0].yhigh
        detector.zmax = usr.detector[0].zhigh

        # TODO read detector type
        detector.dettyp = SHDetType.unknown

        detector.data = np.array(fdata)
        if nscale != 1:
            detector.data *= nscale
            # 1 gigaelectron volt / gram = 1.60217662 x 10-7 Gy
            detector.data *= 1.60217662e-7

        # set units : detector.units are [x,y,z,v,data,detector_title]
        detector.units = [""] * 9

        detector.title = usr.title.decode('ascii')



from collections import namedtuple, defaultdict
import logging
import os

import numpy as np
from enum import IntEnum

from pymchelper.readers.fluka import FlukaBinaryReader
from pymchelper.readers.shieldhit import SHTextReader, SHBinaryReader
from pymchelper.shieldhit.detector.detector_type import SHDetType
from pymchelper.shieldhit.detector.estimator_type import SHGeoType
from pymchelper.writers.excel import ExcelWriter
from pymchelper.writers.plots import ImageWriter, GnuplotDataWriter, PlotDataWriter
from pymchelper.writers.shieldhit import TxtWriter
from pymchelper.writers.sparse import SparseWriter
from pymchelper.writers.trip98 import TripCubeWriter, TripDddWriter

logger = logging.getLogger(__name__)


class Converters(IntEnum):
    """
    Available converters
    """
    txt = 0
    plotdata = 1
    gnuplot = 2
    image = 3
    tripcube = 4
    tripddd = 5
    excel = 6
    sparse = 7


class ErrorEstimate(IntEnum):
    none = 0
    stderr = 1
    stddev = 2


class Axis(IntEnum):
    """
    Axis numbers
    """
    x = 0
    y = 1
    z = 2


_converter_mapping = {
    Converters.txt: TxtWriter,
    Converters.gnuplot: GnuplotDataWriter,
    Converters.plotdata: PlotDataWriter,
    Converters.image: ImageWriter,
    Converters.tripcube: TripCubeWriter,
    Converters.tripddd: TripDddWriter,
    Converters.excel: ExcelWriter,
    Converters.sparse: SparseWriter
}


class Detector:
    """
    Holds data read from single estimator
    """
    data = None
    error = None

    nstat = -1

    xmin = float("NaN")
    xmax = float("NaN")
    nx = -1

    ymin = float("NaN")
    ymax = float("NaN")
    ny = -1

    zmin = float("NaN")
    zmax = float("NaN")
    nz = -1

    dettyp = SHDetType.unknown
    geotyp = SHGeoType.unknown
    particle = 0

    # number of files
    counter = -1

    _M2 = None  # accumulator needed by average_with_other method

    def read(self, filename, nscale=1):
        """
        Reads binary file with. Automatically discovers which reader should be used.
        :param filename: binary file name
        :param nscale:
        :return: none
        """
        reader = SHTextReader(filename)

        # check if binary file is generated with SHIELD-HIT12A version > 0.6
        #  (in that case it may or may not have .bdo extension)
        # this check will also pass is file is generated with older SHIELD-HIT12A version
        #  (in that case we rely on the file extension)
        if SHBinaryReader(filename).test_version_0p6() or filename.endswith((".bdo", ".bdox")):
            reader = SHBinaryReader(filename)
        # find better way to discover if file comes from Fluka
        elif "_fort" in filename:
            reader = FlukaBinaryReader(filename)
        reader.read(self, nscale)
        self.counter = 1

    def average_with_nan(self, other_detectors, error_estimate=ErrorEstimate.stderr):
        """
        Average (not add) data with other detector, excluding malformed data (NaN) from averaging.
        :param other_detectors:
        :param error_estimate:
        :return:
        """
        # TODO add compatibility check
        _l = [det.data for det in other_detectors]
        _l.append(self.data)
        self.counter += len(other_detectors)
        self.nstat += sum(det.nstat for det in other_detectors)
        self.data = np.nanmean(_l, axis=0)
        if error_estimate != ErrorEstimate.none:
            # s = stddev = sqrt(1/(n-1)sum(x-<x>)**2)
            # s : corrected sample standard deviation
            self.error = np.nanstd(_l, axis=0, ddof=1)

    def average_with_other(self, other_detector, error_estimate=ErrorEstimate.stderr):
        """
        Average (not add) data with other detector
        :param other_detector:
        :param error_estimate:
        :return:
        """

        # Running variance algorithm based on algorithm by B. P. Welford,
        # presented in Donald Knuth's Art of Computer Programming, Vol 2, page 232, 3rd edition.
        # Can be found here: http://www.johndcook.com/blog/standard_deviation/
        # and https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Online_algorithm
        self.counter += 1
        self.nstat += other_detector.nstat
        delta = other_detector.data - self.data                    # delta = x - mean
        self.data += delta / self.counter                          # mean += delta / n
        if error_estimate != ErrorEstimate.none:
            self._M2 += delta * (other_detector.data - self.data)  # M2 += delta * (x - mean)

            # unbiased sample variance is stored in `self._M2 / (self.counter - 1)`
            # unbiased sample standard deviation in classical algorithm is calculated as (sqrt(1/(n-1)sum(x-<x>)**2)
            # here it is calculated as square root of unbiased sample variance:
            self.error = np.sqrt(self._M2 / (self.counter - 1))

    def save(self, filename, options):
        """
        Save data to the file, using list of converters
        :param filename:
        :param options:
        :return:
        """
        writer = _converter_mapping[Converters[options.command]](filename, options)
        logger.info("Writing file with corename {:s}".format(filename))
        writer.write(self)

    def __str__(self):
        result = ""
        result += "data" + str(self.data[0].shape) + "\n"
        if self.error is not None:
            result += "error" + str(self.error[0].shape) + "\n"
        result += "nstat = {:d}\n".format(self.nstat)
        result += "X {:g} - {:g} ({:d} items)\n".format(self.xmin, self.xmax, self.nx)
        result += "Y {:g} - {:g} ({:d} items)\n".format(self.ymin, self.ymax, self.ny)
        result += "Z {:g} - {:g} ({:d} items)\n".format(self.zmin, self.zmax, self.nz)
        result += "dettyp {:s}\n".format(self.dettyp.name)
        result += "counter of files {:d}\n".format(self.counter)
        result += "dimension {:d}\n".format(self.dimension)
        result += "Xs {:d}\n".format(len(list(self.x)))
        result += "Xp {:d}\n".format(len(list(self.axis_values(Axis.x, plotting_order=True))))
        result += "Ys {:d}\n".format(len(list(self.y)))
        result += "Yp {:d}\n".format(len(list(self.axis_values(Axis.y, plotting_order=True))))
        result += "Zs {:d}\n".format(len(list(self.z)))
        result += "Zp {:d}\n".format(len(list(self.axis_values(Axis.z, plotting_order=True))))
        result += "V {:d}\n".format(len(list(self.data)))
        return result

    @staticmethod
    def _running_index_i(p, _, k_max):
        return p % k_max

    @staticmethod
    def _running_index_j(p, j_max, k_max):
        return (p // k_max) % j_max

    @staticmethod
    def _running_index_k(p, j_max, k_max):
        return p // (j_max * k_max)

    def _running_indices(self, axis_no):
        if axis_no == 0:
            return self._running_index_i
        elif axis_no == 1:
            return self._running_index_j
        elif axis_no == 2:
            return self._running_index_k
        return None

    @staticmethod
    def _linspace(tmin, tmax, tn, ind):
        dt = (tmax - tmin) / tn
        return tmin + (ind + 0.5) * dt

    def axis_values(self, axis_no, plotting_order=False):
        if plotting_order:
            axis_no = self._axes_plotting_order[axis_no]
        a = self.axis_data(axis_no)
        for p in range(self.nx * self.ny * self.nz):
            ind = self._running_indices(axis_no)(p, self.ny, self.nx)
            yield self._linspace(a.min, a.max, a.n, ind)

    @property
    def x(self):
        return self.axis_values(Axis.x)

    @property
    def y(self):
        return self.axis_values(Axis.y)

    @property
    def z(self):
        return self.axis_values(Axis.z)

    @property
    def v(self):
        return self.data

    @property
    def e(self):
        return self.error

    AxisData = namedtuple('AxisData', ['min', 'max', 'n', 'number'])

    def axis_data(self, axis_number, plotting_order=False):
        if plotting_order:
            axis_number = self._axes_plotting_order[axis_number]
        if axis_number == Axis.x:
            return self.AxisData(min=self.xmin, max=self.xmax, n=self.nx, number=axis_number)
        elif axis_number == Axis.y:
            return self.AxisData(min=self.ymin, max=self.ymax, n=self.ny, number=axis_number)
        elif axis_number == Axis.z:
            return self.AxisData(min=self.zmin, max=self.zmax, n=self.nz, number=axis_number)
        else:
            return None

    @property
    def is_valid(self):
        valid_counters = self.nx > 0 and self.ny > 0 and self.nz > 0
        data_exists = self.data is not None
        borders_correct = self.xmax >= self.xmin and\
            self.ymax >= self.ymin and\
            self.zmax >= self.zmin
        nstat_correct = self.nstat > 0
        return valid_counters and data_exists and borders_correct and nstat_correct

    # 0,1,2,3
    @property
    def dimension(self):
        if self.is_valid:
            return 3 - (self.nx, self.ny, self.nz).count(1)
        else:
            return -1

    @property
    def _axes_plotting_order(self):
        result = (Axis.x, Axis.y, Axis.z)
        if self.dimension == 1:
            if self.nx > 1:
                result = (Axis.x, Axis.y, Axis.z)  # X variable; Y,Z constant
            elif self.ny > 1:
                result = (Axis.y, Axis.x, Axis.z)  # Y variable; X,Z constant
            elif self.nz > 1:
                result = (Axis.z, Axis.x, Axis.y)  # Z variable; X,Y constant
        elif self.dimension == 2:
            if self.nx == 1:
                result = (Axis.y, Axis.z, Axis.x)  # Y,Z variable; X constant
            elif self.ny == 1:
                result = (Axis.x, Axis.z, Axis.y)  # X,Z variable; Y constant
            elif self.nz == 1:
                result = (Axis.x, Axis.y, Axis.z)  # X,Y variable; Z constant
        return result


def merge_list(input_file_list,
               output_file,
               options):
    """
    Takes set of input file names, containing data from the same estimator.
    All input files are read and data is filled (and summed) into detector structure.
    Finally data stored in detector is averaged and saved to output file.
    :param input_file_list: list of input files
    :param output_file: name of output file
    :param options: list of parsed options
    :return: none
    """
    first = Detector()
    first.read(input_file_list[0], options.nscale)

    other_detectors = []

    # allocate memory for accumulator needed in standard deviation calculation
    # not needed if:
    #  - averaging is done ignoring NaNs (then numpy nanvar function is used)
    #  - processing only one file
    #  - user requested not to include errors
    if not options.nan and len(input_file_list) > 1 and options.error != ErrorEstimate.none:
        first._M2 = np.zeros_like(first.data)

    # set errors to zero also if reading single file
    if options.error != ErrorEstimate.none:
        first.error = np.zeros_like(first.data)

    # loop over second and next files, if present
    for file in input_file_list[1:]:
        next_one = Detector()
        next_one.read(file, options.nscale)
        if options.nan:
            other_detectors.append(next_one)  # read all detector files into memory
        else:
            first.average_with_other(other_detector=next_one, error_estimate=options.error)

    # user requested averaging ignoring nan and more than one file are present
    if other_detectors and options.nan:
        first.average_with_nan(other_detectors, error_estimate=options.error)

    # up to now first.error stores standard deviation
    # if user requested standard error then we calculate it as:
    # S = stderr = stddev / sqrt(n), or in other words,
    # S = s/sqrt(N) where S is the corrected standard deviation of the mean.
    if len(input_file_list) > 1 and options.error == ErrorEstimate.stderr:
        first.error /= np.sqrt(first.counter)  # np.sqrt() always returns np.float64

    if output_file is None:
        output_file = input_file_list[0][:-4]

    output_dir = os.path.dirname(output_file)
    if output_dir:  # output directory has been found, output_file is not a plain file in current dir
        if not os.path.exists(output_dir):  # directory doesn't exists
            os.makedirs(output_dir)  # let us create it
    first.save(output_file, options)


def _process_one_group(core_name, group_with_same_core, outputdir, options):
    core_dirname, core_basename = os.path.split(core_name)
    if outputdir is None:
        output_file = os.path.join(core_dirname, core_basename)
    else:
        output_file = os.path.join(outputdir, core_basename)
    logger.debug("Setting output core name " + output_file)
    merge_list(group_with_same_core, output_file, options)


def merge_many(input_file_list,
               outputdir,
               options,
               jobs):
    """
    Takes set of input file names, belonging to possibly different estimators.
    Input files are grouped according to the estimators and for each group
    merging is performed, as in @merge_list method.
    Output file name is automatically generated.
    :param input_file_list: list of input files
    :param outputdir: output directory
    :param options: list of parsed options
    :param jobs: number of CPU cores to use (-1 means all)
    :return: none
    """
    core_names_dict = defaultdict(list)
    # keys - core_name, value - list of full paths to corresponding files

    # loop over input list of file paths
    for filepath in input_file_list:

        # extract basename for inspection
        basename = os.path.basename(filepath)

        # SHIELD-HIT12A binary file encountered
        if SHBinaryReader(filepath).test_version_0p6() or filepath.endswith(('.bdo', '.bdox')):
            # we expect the basename to follow one of two conventions:
            #  - corenameABCD.bdo (where ABCD is 4-digit integer)
            #  - corename.bdo
            core_name = basename[:-4]  # assume no number in the basename
            if basename[-8:-4].isdigit() and len(basename[-8:-4]) == 4:  # check if number present
                core_name = basename[:-8]
            core_names_dict[core_name].append(filepath)
        elif "_fort." in filepath:  # Fluka binary file encountered
            core_name = filepath[-2:]
            core_names_dict[core_name].append(filepath)

    # parallel execution of output file generation, using all CPU cores
    # see http://pythonhosted.org/joblib
    try:
        from joblib import Parallel, delayed
        logger.info("Parallel processing on {:d} jobs (-1 means all)".format(jobs))
        # options.verbose count the number of `-v` switches provided by user
        # joblib Parallel class expects the verbosity as a larger number (i.e. multiple of 10)
        worker = Parallel(n_jobs=jobs, verbose=options.verbose * 10)
        worker(
            delayed(_process_one_group)(core_name, group_with_same_core, outputdir, options)
            for core_name, group_with_same_core in core_names_dict.items()
        )
    except (ImportError, SyntaxError):
        # single-cpu implementation, in case joblib library fails (i.e. Python 3.2)
        logger.info("Single CPU processing")
        for core_name, group_with_same_core in core_names_dict.items():
            _process_one_group(core_name, group_with_same_core, outputdir, options)



import os
import logging

import numpy as np

logger = logging.getLogger(__name__)


class PlotDataWriter:
    def __init__(self, filename, options):
        self.filename = filename
        if not self.filename.endswith(".dat"):
            self.filename += ".dat"

    def write(self, detector):
        logger.info("Writing: " + self.filename)

        data = np.array(detector.data)
        error = np.array(detector.error)

        # change units for LET from MeV/cm to keV/um
        from pymchelper.shieldhit.detector.detector_type import SHDetType
        if detector.dettyp in (SHDetType.dlet, SHDetType.dletg, SHDetType.tlet, SHDetType.tletg):
            data *= np.float64(0.1)  # 1 MeV / cm = 0.1 keV / um
            if np.any(error):
                error *= np.float64(0.1)  # 1 MeV / cm = 0.1 keV / um

        axis_data_column = [list(detector.axis_values(i, plotting_order=True)) for i in range(detector.dimension)]

        fmt = "%g" + " %g" * detector.dimension
        data_to_save = axis_data_column + [data.ravel()]  # ravel needed to change arrays like [[1]] to [1]

        # if error information is present save it as additional column
        if detector.error is not None:
            fmt += " %g"
            data_to_save += [error.ravel()]

        # transpose from rows to columns
        data_columns = np.transpose(data_to_save)

        # save space-delimited text file
        np.savetxt(self.filename, data_columns, fmt=fmt, delimiter=' ')


class GnuplotDataWriter:
    def __init__(self, filename, options):
        self.data_filename = filename
        self.script_filename = filename
        self.plot_filename = filename

        if not self.plot_filename.endswith(".png"):
            self.plot_filename += ".png"
        if not self.script_filename.endswith(".plot"):
            self.script_filename += ".plot"
        if not self.data_filename.endswith(".dat"):
            self.data_filename += ".dat"

        dirname = os.path.split(self.script_filename)[0]
        self.awk_script_filename = os.path.join(dirname, "addblanks.awk")

    _awk_2d_script_content = """/^[[:blank:]]*#/ {next} # ignore comments (lines starting with #)
NF < 3 {next} # ignore lines which don't have at least 3 columns
$2 != prev {printf \"\\n\"; prev=$2} # print blank line
{print} # print the line
    """

    _header = """set term png
set output \"{plot_filename}\"
set title \"{title}\"
set xlabel \"{xlabel}\"
set ylabel \"{ylabel}\"
"""

    _error_plot_command = "'./{data_filename}' u 1:(max($2-$3,0.0)):($2+$3) w filledcurves " \
                          "fs transparent solid 0.2 lc 3 title '1-sigma confidence', "

    _plotting_command = {
        1: """max(x,y) = (x > y) ? x : y
plot {error_plot} './{data_filename}' u 1:2 w l lt 1 lw 2 lc -1 title 'mean value'
        """,
        2: """set view map
splot \"<awk -f addblanks.awk '{data_filename}'\" u 1:2:3 with pm3d
"""
    }

    def write(self, detector):
        # skip plotting 0-D and 3-D data
        if detector.dimension in (0, 3):
            return

        # set labels
        x_axis_number = detector.axis_data(0, plotting_order=True).number
        xlabel = detector.units[x_axis_number]
        if detector.dimension == 1:
            ylabel = ImageWriter._make_label(detector.units[4], detector.title)
        elif detector.dimension == 2:
            y_axis_number = detector.axis_data(1, plotting_order=True).number
            ylabel = detector.units[y_axis_number]

            # for 2-D plots writte additional awk script to convert data
            # as described in gnuplot faq: http://www.gnuplot.info/faq/faq.html#x1-320003.9
            with open(self.awk_script_filename, 'w') as script_file:
                logger.info("Writing: " + self.awk_script_filename)
                script_file.write(self._awk_2d_script_content)

        # save gnuplot script
        with open(self.script_filename, 'w') as script_file:
            logger.info("Writing: " + self.script_filename)
            script_file.write(self._header.format(plot_filename=self.plot_filename, xlabel=xlabel, ylabel=ylabel,
                                                  title=detector.title))
            plt_cmd = self._plotting_command[detector.dimension]

            # add error plot if error data present
            err_cmd = ""
            if np.any(detector.error):
                err_cmd = self._error_plot_command.format(data_filename=self.data_filename)

            script_file.write(plt_cmd.format(data_filename=self.data_filename, error_plot=err_cmd))


class ImageWriter:
    def __init__(self, filename, options):
        self.plot_filename = filename
        if not self.plot_filename.endswith(".png"):
            self.plot_filename += ".png"
        self.colormap = options.colormap

    default_colormap = 'gnuplot2'

    @staticmethod
    def _make_label(unit, name):
        return name + " " + "[" + unit + "]"

    def _save_2d_error_plot(self, detector, xlist, ylist, elist):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt

        plt.pcolormesh(xlist, ylist, elist.clip(0.0), cmap=self.colormap)
        y_axis_number = detector.axis_data(1, plotting_order=True).number
        y_axis_name = detector.units[6 + y_axis_number]
        plt.ylabel(self._make_label(detector.units[y_axis_number], y_axis_name))
        cbar = plt.colorbar()
        cbar.set_label(detector.units[4], rotation=270, verticalalignment='bottom')
        base_name, _ = os.path.splitext(self.plot_filename)
        plt.savefig(base_name + "_error.png")
        plt.close()

    def write(self, detector):
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt

        # skip plotting 0-D and 3-D data
        if detector.dimension in (0, 3):
            return

        data = np.array(detector.data)
        error = np.array(detector.error)

        # change units for LET from MeV/cm to keV/um
        from pymchelper.shieldhit.detector.detector_type import SHDetType
        if detector.dettyp in (SHDetType.dlet, SHDetType.dletg, SHDetType.tlet, SHDetType.tletg):
            data *= np.float64(0.1)  # 1 MeV / cm = 0.1 keV / um
            if np.any(error):
                error *= np.float64(0.1)  # 1 MeV / cm = 0.1 keV / um

        logger.info("Writing: " + self.plot_filename)

        x_axis_number = detector.axis_data(0, plotting_order=True).number
        x_axis_name = detector.units[6 + x_axis_number]
        plt.xlabel(self._make_label(detector.units[x_axis_number], x_axis_name))
        xlist = list(detector.axis_values(0, plotting_order=True))  # make list of values from generator

        # 1-D plotting
        if detector.dimension == 1:

            # add optional error area
            if np.any(detector.error):
                plt.fill_between(xlist,
                                 (data - error).clip(0.0),
                                 (data + error).clip(0.0, 1.05 * (detector.v.max())),
                                 alpha=0.2, edgecolor='#CC4F1B', facecolor='#FF9848', antialiased=True)
            plt.ylabel(self._make_label(detector.units[4], detector.title))
            plt.plot(xlist, data)
        elif detector.dimension == 2:
            ylist = list(detector.axis_values(1, plotting_order=True))   # make list of values from generator

            xn = detector.axis_data(0, plotting_order=True).n
            yn = detector.axis_data(1, plotting_order=True).n

            shape_tuple = (yn, xn)
            xlist = np.asarray(xlist).reshape(shape_tuple)
            ylist = np.asarray(ylist).reshape(shape_tuple)
            zlist = data.reshape(shape_tuple)

            # add error plot if error data present
            if np.any(detector.error):
                elist = error.reshape(shape_tuple)
                self._save_2d_error_plot(detector, xlist, ylist, elist)

            x_axis_number = detector.axis_data(0, plotting_order=True).number
            x_axis_name = detector.units[6 + x_axis_number]
            plt.xlabel(self._make_label(detector.units[x_axis_number], x_axis_name))

            y_axis_number = detector.axis_data(1, plotting_order=True).number
            y_axis_name = detector.units[6 + y_axis_number]
            plt.ylabel(self._make_label(detector.units[y_axis_number], y_axis_name))

            plt.pcolormesh(xlist, ylist, zlist, cmap=self.colormap)
            cbar = plt.colorbar()
            cbar.set_label(detector.units[4], rotation=270, verticalalignment='bottom')
        plt.savefig(self.plot_filename)
        plt.close()



#!/usr/bin/env python

import glob
import logging
import sys

import argparse

from pymchelper.detector import merge_list, merge_many, Converters, ErrorEstimate
from pymchelper.writers.plots import ImageWriter

logger = logging.getLogger(__name__)


def set_logger_level(args):
    if args.quiet:
        if args.quiet == 1:
            level = "WARNING"
        if args.quiet == 2:
            level = "ERROR"
        else:
            level = "CRITICAL"
    elif args.verbose:
        level = "DEBUG"
    else:
        level = "INFO"
    logging.basicConfig(level=level)


def add_default_options(parser):
    import pymchelper
    parser.add_argument('input', help='input filename, file list or pattern', type=str)
    parser.add_argument('output', help='output filename or directory', nargs='?')
    parser.add_argument('-j', '--jobs', help='number of parallel jobs to use (-1 means all CPUs)', default=-1, type=int)
    parser.add_argument('--many', help='automatically merge data from various sources', action="store_true")
    parser.add_argument('-a', '--nan', help='ignore NaN in averaging', action="store_true")
    parser.add_argument('-e', '--error',
                        help='type of error estimate to add (default: ' + ErrorEstimate.stderr.name + ')',
                        default=ErrorEstimate.stderr.name,
                        choices=[x.name for x in ErrorEstimate],
                        type=str)
    parser.add_argument('-n', '--nscale', help='scale with number of primaries N.', default=1, type=float)
    parser.add_argument('-v',
                        '--verbose',
                        action='count',
                        default=0,
                        help='give more output. Option is additive, and can be used up to 3 times')
    parser.add_argument('-q', '--quiet', action='count', default=0, help='be silent')
    parser.add_argument('-V', '--version', action='version', version=pymchelper.__version__)


def main(args=sys.argv[1:]):
    import pymchelper
    import os

    _progname = os.path.basename(sys.argv[0])
    _helptxt = 'Universal converter for FLUKA and SHIELD-HIT12A generated files.'
    _epitxt = '''Type '{:s} <converter> --help' for help on a specific converter.'''.format(_progname)

    parser = argparse.ArgumentParser(description=_helptxt, epilog=_epitxt)

    # subparsers = parser.add_subparsers(title='available converters', metavar='...')
    subparsers = parser.add_subparsers(dest='command', metavar='converter')

    parser_txt = subparsers.add_parser(Converters.txt.name, help='converts to plain txt file')
    add_default_options(parser_txt)

    parser_image = subparsers.add_parser(Converters.image.name, help='converts to PNG images')
    add_default_options(parser_image)
    parser_image.add_argument("--colormap",
                              help='image color map, see http://matplotlib.org/users/colormaps.html '
                                   'for list of possible options (default: ' + ImageWriter.default_colormap + ')',
                              default=ImageWriter.default_colormap,
                              type=str)

    parser_excel = subparsers.add_parser(Converters.excel.name, help='converts to MS Excel file')
    add_default_options(parser_excel)

    parser_plotdata = subparsers.add_parser(Converters.plotdata.name, help='converts to gnuplot data')
    add_default_options(parser_plotdata)

    parser_gnuplot = subparsers.add_parser(Converters.gnuplot.name, help='converts to gnuplot script')
    add_default_options(parser_gnuplot)

    parser_sparse = subparsers.add_parser(Converters.sparse.name, help='converts to sparse matrix format')
    add_default_options(parser_sparse)
    parser_sparse.add_argument("--threshold",
                               help='only values greater than threshold are saved',
                               type=float,
                               default=0.0)

    parser_tripcube = subparsers.add_parser(Converters.tripcube.name, help='converts to trip98 data cube')
    add_default_options(parser_tripcube)

    parser_tripddd = subparsers.add_parser(Converters.tripddd.name, help='converts to trip98 ddd file')
    add_default_options(parser_tripddd)
    parser_tripddd.add_argument("--energy",
                                help='energy of the beam [MeV/amu]',
                                type=float)
    parser_tripddd.add_argument("--ngauss",
                                help='number of Gauss curves to fit (default: 2)',
                                choices=(0, 1, 2),
                                default=2,
                                type=int)

    parser.add_argument('-V', '--version', action='version', version=pymchelper.__version__)

    parsed_args = parser.parse_args(args)

    if parsed_args.command is not None:
        set_logger_level(parsed_args)

        # check if output directory exists
        if parsed_args.output is not None:
            output_dir = os.path.dirname(parsed_args.output)
            if output_dir and not os.path.exists(output_dir):
                raise IOError("Directory {}/ does not exist.".format(output_dir))

        # TODO add filename discovery
        files = sorted(glob.glob(parsed_args.input))
        if not files:
            logger.error('File does not exist: ' + parsed_args.input)

        parsed_args.error = ErrorEstimate[parsed_args.error]

        # check required options for tripddd parser
        if parsed_args.command == Converters.tripddd.name and not parsed_args.energy:
            logger.error("Option --energy is required, provide an energy value")
            return 1

        if parsed_args.many:
            merge_many(files, parsed_args.output, parsed_args, parsed_args.jobs)
        else:
            merge_list(files, parsed_args.output, parsed_args)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
